from argparse import ArgumentParser
from os.path import abspath, split
from os import makedirs
from datetime import datetime
from calendar import month_name

import numpy as np

from rad_tools.io.tb2j import read_exchange_model
from rad_tools.routines import OK, RESET, YELLOW


def manager(output_name="template",
            input_filename=None,
            R_vector=None,
            max_distance=None,
            min_distance=None,
            distance=None):
    r"""
    Main function call of the rad-make-template.py script.

    Full documentation on the behaviour is available
    :ref:`here <rad-make-template>`. Parameters of the function directly
    corresponds to the arguments of the script.

    If you want to have the behaviour of the rad-make-template.py script
    but in a format of a fuction call use this function.
    """

    if distance is not None:
        min_distance = distance
        max_distance = distance

    if R_vector is not None:
        R_vector = np.array(R_vector[:len(R_vector) // 3 * 3],
                            dtype=int).reshape((len(R_vector)//3, 3))
        R_vector = list(map(tuple, R_vector.tolist()))

    try:
        makedirs(split(output_name)[0])
    except FileExistsError:
        pass
    except FileNotFoundError:
        pass

    template = ('=' * 20 + "\n" +
                "Neighbors template:\n" +
                "i j R_a R_b R_c\n" +
                '-' * 20 + "\n" +
                "J1 $J_1$\n" +
                "atom1 atom2  0  0  0\n" +
                "atom1 atom2  1  0  0\n" +
                "atom1 atom1 -1  0  2\n" +
                '-' * 20 + "\n" +
                "J2\n" +
                "atom2 atom1  9  5 -3\n" +
                "atom1 atom2  1  4  0\n" +
                "atom2 atom2  1  0  2\n" +
                '=' * 20 + "\n")
    cd = datetime.now()
    with open(f"{output_name}.txt", "w") as file:
        if input_filename is None:
            file.write(
                f"Template is created " +
                f"on {cd.day} {month_name[cd.month]} {cd.year}" +
                f" at {cd.hour}:{cd.minute}:{cd.second} by rad-tools\n\n")

            file.write(template)
        else:
            file.write(
                f"Template is created based on the file: {input_filename}\n" +
                f"on {cd.day} {month_name[cd.month]} {cd.year}" +
                f" at {cd.hour}:{cd.minute}:{cd.second} by rad-tools\n\n")

            model = read_exchange_model(input_filename)
            model.filter(min_distance=min_distance,
                         max_distance=max_distance,
                         R_vector=R_vector)
            file.write('=' * 20 + "\n" +
                       "Neighbors template:\n" +
                       "i j R_a R_b R_c\n" +
                       '-' * 20 + "\n" + "Name placeholder" + "\n")
            for atom1, atom2, R in model.bonds:
                file.write(f"{atom1:4} {atom2:4} " +
                           f"{R[0]:3.0f} {R[1]:3.0f} {R[2]:3.0f}\n")
            file.write('=' * 20 + "\n")
    print(f"{OK}Template draft is in " +
          f"{abspath(output_name)}.txt{RESET}")
    print(f"{YELLOW}Do not forget to correct the template draft " +
          f"to your needs!{RESET}")


def get_parser():
    parser = ArgumentParser(
        description="Script for the creation of template`s template")

    parser.add_argument("-on", "--output-name",
                        type=str,
                        default='template',
                        help="""
                        Relative or absolute path to the template output file.
                        """
                        )
    parser.add_argument("-if", "--input-filename",
                        type=str,
                        default=None,
                        help="""
                        Relative or absulute path to the *exchange.out* file,
                        including the name and extention of the file itself.
                        """
                        )
    parser.add_argument("-R", "--R-vector",
                        type=int,
                        nargs="*",
                        default=None,
                        help="""
                        R vectors for filtering the model.
                        """
                        )
    parser.add_argument("-maxd", "--max-distance",
                        type=float,
                        default=None,
                        help="""
                        (<=) Maximum distance.
                        """
                        )
    parser.add_argument("-mind", "--min-distance",
                        type=float,
                        default=None,
                        help="""
                        (>=) Minimum distance.
                        """
                        )
    parser.add_argument("-d", "--distance",
                        type=float,
                        default=None,
                        help="""
                        (=) Exact distance.
                        """
                        )

    return parser



#! /usr/local/bin/python3

from argparse import ArgumentParser
from os import makedirs, walk
from os.path import abspath, join
from copy import deepcopy

from rad_tools.routines import OK, RESET, YELLOW

import matplotlib.pyplot as plt
from matplotlib import __version__ as matplotlib_version
import numpy as np

CASES = ["collinear, spin-unpolarized",
         "collinear, spin-polarized",
         "non-collinear, non spin-orbit"]


def analyse_input_folder(input_path, filpdos=None):
    r"""
    Analyse input folder, detects seednames for the dos output files.

    Parameters
    ----------
    input_path : str
        Directory with DOS files.
    filpdos : str, default None
        Seddname for the DOS file. If not provided then all seednames
        from the folder are detected and processed.

    Returns
    -------
    filpdos : list
        List of seednames found in ``input_path``.
    filenames : list
        List of lists of filenames for each seedname in ``filpdos``.
    """

    files = []
    for (dirpath, dirnames, filenames) in walk(input_path):
        files.extend(filenames)
        break

    filenames = []
    if filpdos is not None:
        for filename in files:
            if filpdos in filename:
                filenames.append(filename)
        files = filenames
        filpdos = [filpdos]
        filenames = [filenames]
    else:
        filpdos = set()
        for filename in files:
            if ".pdos_tot" in filename and ".pdos_tot" == filename[-9:]:
                filpdos.add(filename[:-9])
                filenames.append(filename)
            elif ".pdos_atm#" in filename:
                filpdos.add(filename.split(".pdos_atm#")[0])
                filenames.append(filename)
        filpdos = list(filpdos)

        print(f"Following DOS seednames (filpdos) are detected:{YELLOW}")
        for item in filpdos:
            print(f"   * {item}")
        print(RESET)

        if len(filpdos) != 1:
            tmp = [[] for _ in range(0, len(filpdos))]
            for filename in filenames:
                for i, seedname in enumerate(filpdos):
                    if seedname in filename:
                        tmp[i].append(filename)
            filenames = tmp
        else:
            filenames = [filenames]
    return filpdos, filenames


def detect_case(filename):
    r"""
    Detects case of the DOS calculations.

    Parameters
    ----------
    filename: str
        Path to the file in which the calculation case have to be detected

    Returns
    -------
    case : int
        Case index.
    """

    with open(filename) as file:
        header = file.readline().lower()
    case = None
    if ("dos" in header
            and "pdos" in header):
        case = 0
    if ("dos" in header
        and "pdosup" in header
            and "pdosdw" in header):
        case = 2
    if ("dosup" in header
        and "dosdw" in header
        and "pdosup" in header
            and "pdosdw" in header):
        case = 1
    if case is None:
        raise RuntimeError("Unable to detect case, analized header:\n" +
                           f"{header}")
    return case


def decompose_filenames(filenames):
    r"""
    Decompose filenames and extract information about atoms and projections.

    Parameters
    ----------
    filenames : str
        List of filenames with projected DOS (not pdos_total).

    Returns
    -------
    atoms : dist
        Distionary with the atom labels and their numbers.

        .. code-block:: python

            atoms = {atom1: [n_1, n_2, n_3, ...]}
    wfcs : dict
        Dictionary of projection functions and their numbers.

        .. code-block:: python

            wfcs = {(atom1, atom1_number): (wfc_label, wfc_number)}
    """

    atoms = {}
    wfcs = {}
    for filename in filenames:

        # Detect names and numbers
        atom_number = int(filename.split(".pdos_atm#")[1].split("(")[0])
        atom_symbol = filename.split("(")[1].split(")")[0]
        wfc_number = int(filename.split(")_wfc#")[1].split("(")[0])
        wfc_symbol = filename.split(")_wfc#")[1].split("(")[1].split(")")[0]

        if atom_symbol not in atoms:
            atoms[atom_symbol] = []
        atoms[atom_symbol].append(atom_number)

        if (atom_symbol, atom_number) not in wfcs:
            wfcs[(atom_symbol, atom_number)] = []
        wfcs[(atom_symbol, atom_number)].append((wfc_symbol, wfc_number))

    # Sort entries
    for atom in atoms:
        atoms[atom] = list(set(atoms[atom]))
        atoms[atom].sort()
    for key in wfcs:
        wfcs[key].sort(key=lambda x: x[1])

    return atoms, wfcs


def combine_by_atoms(input_path,
                     seedname,
                     output_path,
                     atoms,
                     wfcs,
                     case):
    r"""
    Combine PDOS by atom type.

    Parameters
    ----------
    input_path : str
    seedname : str
    output_path : str
    atoms : dict
    wfcs : dict
    case : int
    """
    dos_by_atom = {}
    print(f"Start to combine PDOS by atoms")
    for atom in atoms:
        print(f"    {len(atoms[atom])} atoms of {atom} detected")
        for wfc_symbol, wfc_number in wfcs[(atom, atoms[atom][0])]:
            with open(join(input_path,
                           f"{seedname}.pdos_atm#{atoms[atom][0]}({atom})_wfc#{wfc_number}({wfc_symbol})")) as file:
                header = file.readline().replace("\n", "")
            for i in atoms[atom]:
                if i == atoms[atom][0]:
                    dos = np.loadtxt(join(input_path,
                                          f"{seedname}.pdos_atm#{i}({atom})_wfc#{wfc_number}({wfc_symbol})"),
                                     skiprows=1).T
                else:
                    dos += np.loadtxt(join(input_path,
                                           f"{seedname}.pdos_atm#{i}({atom})_wfc#{wfc_number}({wfc_symbol})"),
                                      skiprows=1).T
            dos[0] = dos[0] / len(atoms[atom])
            energy = dos[0]
            if case == 0:
                dos_by_atom[f"{atom}, {wfc_symbol}"] = dos[1]
            elif case == 1 or case == 2:
                dos_by_atom[f"{atom}, {wfc_symbol}"] = [dos[1], dos[2]]

            np.savetxt(join(output_path,
                            seedname,
                            "summed-by-atom",
                            f"{atom}_wfc#{wfc_number}({wfc_symbol})"),
                       dos.T,
                       header=header,
                       comments="",
                       fmt="%8.3f  "+"%.3E  "*(dos.shape[0]-1))
    print(f"Files with PDOS combined by atom are in " +
          f"{abspath(join(output_path, seedname, 'summed-by-atom'))}")


def filter_window(dos, window, efermi=0.):
    r"""
    Filter dos with respect to the window and make a shift to Fermi energy.

    Parameters
    ----------
    dos : array
        Density of states array.
    window : tuple, default None
        Limits for the energy.
    efermi : float, default 0
        Fermi energy.
    """

    dos[0] -= efermi
    if window is not None:
        i_min, i_max = 0, 0
        for i in range(0, len(dos[0])):
            if dos[0][i] < window[0]:
                i_min = i
            if dos[0][i] < window[1]:
                i_max = i
        dos = dos.T[i_min:i_max].T
    return dos


def plot_pdos_tot(dos, output_name, case, window=None, efermi=0.):

    dos = filter_window(dos, window=window, efermi=efermi)

    ax = plt.subplots(figsize=(8, 4))[1]

    ax.set_xlabel("Energy, eV", fontsize=18)
    ax.set_ylabel("DOS, states/eV", fontsize=18)
    if efermi != 0:
        ax.set_title(f"DOS and PDOS (0 is Fermi energy)", fontsize=18)
    else:
        ax.set_title(
            f"DOS and PDOS (0 is 0)", fontsize=18)
    ax.vlines(0, 0, 1, transform=ax.get_xaxis_transform(),
              color='grey', linewidths=0.5,
              linestyles='dashed')

    if case == 0:
        ax.fill_between(dos[0], 0, dos[1],
                        lw=0, color='grey',
                        alpha=0.3, label="DOS")
        ax.plot(dos[0], dos[2], "-",
                lw=1, color="orange",
                alpha=0.7, label="PDOS")
        ncol = 1
    if case == 1:
        ax.fill_between(dos[0], 0, dos[1],
                        lw=0, color='blue',
                        alpha=0.3, label="DOS (up)")
        ax.plot(dos[0], dos[3], "-",
                lw=1, color="blue",
                alpha=0.7, label="PDOS (up)")
        ax.fill_between(dos[0], 0, -dos[2],
                        lw=0, color='red',
                        alpha=0.3,  label="DOS (down)")
        ax.plot(dos[0], -dos[4], "-",
                lw=1, color="red",
                alpha=0.7, label="PDOS (down)")
        ncol = 2
    if case == 2:
        ax.fill_between(dos[0], 0, dos[1],
                        lw=0, color='grey',
                        alpha=0.3, label="DOS")
        ax.plot(dos[0], dos[2], "-",
                lw=1, color="blue",
                alpha=0.7, label="PDOS (up)")
        ax.plot(dos[0], -dos[3], "-",
                lw=1, color="red",
                alpha=0.7, label="PDOS (down)")
        ncol = 1

    ax.legend(loc="best", ncol=ncol)

    png_path = f"{output_name}.png"
    plt.savefig(png_path, dpi=600, bbox_inches="tight")
    print(f"Total dos plot  is in {abspath(png_path)}")
    plt.close()


def plot_projected(l, dos, output_name, title, case,
                   window=None, efermi=0.,
                   labels=None, factor=None):
    if l in ["s", "p", "d"]:
        labels = {"s": ["s"],
                  "p": ["$p_z$", "$p_y$", "$p_x$"],
                  "d": ["$d_{z^2}$", "$d_{zx}$",
                        "$d_{zy}$", "$d_{x^2 - y^2}$", "$d_{xy}$"]}
        labels = labels[l]

        factor = {"s": 1,
                  "p": 3,
                  "d": 5}
        factor = factor[l]
    elif labels is None or factor is None:
        raise ValueError
    dos = filter_window(dos, window=window, efermi=efermi)
    if window is None:
        window = (np.amin(dos[0]), np.amax(dos[0]))
    fig, axs = plt.subplots(factor, 1, figsize=(9, factor*2))
    if len(labels) == 1:
        axs = [axs]
    fig.subplots_adjust(hspace=0)
    for i in range(0, factor):
        ax = axs[i]
        ax.set_ylabel('DOS, states/eV', fontsize=12)
        if i == factor - 1:
            ax.set_xlabel('E, ev', fontsize=15)
        else:
            ax.axes.get_xaxis().set_visible(False)
        ax.set_xlim(*tuple(window))
        ax.vlines(0, 0, 1, transform=ax.get_xaxis_transform(),
                  color='grey', linewidths=0.5, linestyles='dashed')
        if title is not None and i == 0:
            ax.set_title(title)

        if case == 0:
            ax.fill_between(dos[0], 0, dos[1],
                            lw=0, color='black',
                            alpha=0.3, label="p")
            ax.plot(dos[0], dos[2+i],
                    "-",
                    lw=0.5,
                    color="black",
                    alpha=0.8,
                    label=labels[i])
        elif case in [1, 2]:
            ax.fill_between(dos[0], 0, dos[1],
                            lw=0, color='blue',
                            alpha=0.3, label=f"{l} (up)")
            ax.plot(dos[0], dos[3 + 2*i],
                    "-",
                    lw=0.5,
                    color="blue",
                    alpha=0.8,
                    label=f"{labels[i]} (up)")
            ax.fill_between(dos[0], 0, -dos[2],
                            lw=0, color='red',
                            alpha=0.3, label=f"{l} (down)")
            ax.plot(dos[0], -dos[4 + 2*i],
                    "-",
                    lw=0.5,
                    color="red",
                    alpha=0.8,
                    label=f"{labels[i]} (down)")
        ax.legend(loc=(1.025, 0.2),
                  bbox_transform=ax.transAxes)

    plt.savefig(f"{output_name}.png", dpi=400, bbox_inches="tight")
    plt.close()


def plot_projected_relative(l, dos, output_name, title, case,
                            window=None, efermi=0.,
                            normalize=False,
                            labels=None, factor=None):
    if l in ["s", "p", "d"]:
        labels = {"s": ["s"],
                  "p": ["$p_z$", "$p_y$", "$p_x$"],
                  "d": ["$d_{z^2}$", "$d_{zx}$",
                        "$d_{zy}$", "$d_{x^2 - y^2}$", "$d_{xy}$"]}
        labels = labels[l]

        factor = {"s": 1,
                  "p": 3,
                  "d": 5}
        factor = factor[l]
    elif labels is None or factor is None:
        raise ValueError
    colors = ["#AC92EB", "#ED5564", "#A0D568", "#FFCE54", "#4FC1E8"]
    dos = filter_window(dos, window=window, efermi=efermi)
    if window is None:
        window = (np.amin(dos[0]), np.amax(dos[0]))
    if case == 0:
        fig, axs = plt.subplots(figsize=(9, 4))
        ax1 = axs
    elif case in [1, 2]:
        fig, axs = plt.subplots(2, 1, figsize=(9, 4))
        ax1 = axs[0]
        ax2 = axs[1]
    fig.subplots_adjust(hspace=0)
    if normalize:
        ax1.set_ylabel('PDOS / LDOS', fontsize=12)
        if case in [1, 2]:
            ax2.set_ylabel('PDOS / LDOS', fontsize=12)
    else:
        ax1.set_ylabel('DOS, states/eV', fontsize=12)
        if case in [1, 2]:
            ax2.set_ylabel('DOS, states/eV', fontsize=12)

    ax1.set_xlim(*tuple(window))
    ax1.vlines(0, 0, 1, transform=ax1.get_xaxis_transform(),
               color='grey', linewidths=0.5, linestyles='dashed')

    if case in [1, 2]:
        ax2.set_xlabel('E, ev', fontsize=15)
        ax2.set_xlim(*tuple(window))
        ax2.vlines(0, 0, 1, transform=ax2.get_xaxis_transform(),
                   color='grey', linewidths=0.5, linestyles='dashed')
        ax1.axes.get_xaxis().set_visible(False)
    else:
        ax1.set_xlabel('E, ev', fontsize=15)

    if title is not None:
        ax1.set_title(title)

    if case in [1, 2]:
        y_up = np.zeros(dos[0].shape)
        y_prev_up = np.zeros(dos[0].shape)
        y_down = np.zeros(dos[0].shape)
        y_prev_down = np.zeros(dos[0].shape)
    else:
        y = np.zeros(dos[0].shape)
        y_prev = np.zeros(dos[0].shape)

    for i in range(0, factor):
        if case == 0:
            y += dos[2+i]
            if normalize:
                ax1.fill_between(dos[0], y_prev/dos[1], y/dos[1],
                                 lw=0, color=colors[i],
                                 label=f"{labels[i]}")
            else:
                ax1.fill_between(dos[0], y_prev, y,
                                 lw=0, color=colors[i],
                                 label=f"{labels[i]}")
            y_prev = deepcopy(y)
        elif case in [1, 2]:
            y_up += dos[3+2*i]
            y_down += dos[4+2*i]
            if normalize:
                ax1.fill_between(dos[0], y_prev_up/dos[1], y_up/dos[1],
                                 lw=0, color=colors[i],
                                 label=f"{labels[i]}")
                ax2.fill_between(dos[0], y_prev_down/dos[2], y_down/dos[2],
                                 lw=0, color=colors[i],
                                 label=f"{labels[i]}")
            else:
                ax1.fill_between(dos[0], y_prev_up, y_up,
                                 lw=0, color=colors[i],
                                 label=f"{labels[i]}")
                ax2.fill_between(dos[0], y_prev_down, y_down,
                                 lw=0, color=colors[i],
                                 label=f"{labels[i]}")
            y_prev_up = deepcopy(y_up)
            y_prev_down = deepcopy(y_down)

    if normalize:
        ax1.set_ylim(0, 1)
        ax2.set_ylim(1, 0)
    else:
        ax1.set_ylim(0, None)
        ax2.set_ylim(0, None)
        ax2.invert_yaxis()
    if matplotlib_version >= "3.7.0":
        ax1.legend(loc=(1.025, 0.2),
                   bbox_transform=ax1.transAxes, title="up", reverse=True)
    else:

        ax1.legend(loc=(1.025, 0.2),
                   bbox_transform=ax1.transAxes, title="up")
    ax2.legend(loc=(1.025, 0.2),
               bbox_transform=ax2.transAxes, title="down")

    plt.savefig(f"{output_name}.png", dpi=400, bbox_inches="tight")
    plt.close()


def manager(input_path,
            filpdos=None,
            output_path=".",
            window=None,
            efermi=0.,
            separate=False,
            relative=False,
            normalize=False):

    try:
        makedirs(output_path)
    except FileExistsError:
        pass

    filpdos, files = analyse_input_folder(input_path, filpdos)

    for s_i, seedname in enumerate(filpdos):

        # Preparations
        print(f"{YELLOW}({s_i + 1}/{len(filpdos)}) " +
              f"Start to work with {seedname} seedname{RESET}")

        case = detect_case(join(input_path, f"{seedname}.pdos_tot"))
        print(f"{CASES[case]} case detected.")

        try:
            makedirs(join(output_path, seedname))
        except FileExistsError:
            pass
        try:
            makedirs(join(output_path, seedname, "summed-by-atom"))
        except FileExistsError:
            pass
        try:
            makedirs(join(output_path, seedname,
                          "summed-by-atom", "wfc-in-atoms"))
        except FileExistsError:
            pass
        if separate:
            try:
                makedirs(join(output_path, seedname, "individual-plots"))
            except FileExistsError:
                pass
            try:
                makedirs(join(output_path, seedname,
                              "individual-plots", "wfc-in-atoms"))
            except FileExistsError:
                pass

        atoms, wfcs = decompose_filenames(
            [file for file in files[s_i] if ".pdos_tot" != file[-9:]])

        # Plot PDOS vs DOS
        total_dos = np.loadtxt(join(input_path, f"{seedname}.pdos_tot"),
                               skiprows=1).T
        plot_pdos_tot(total_dos,
                      join(output_path, seedname, "pdos-vs-dos"),
                      case,
                      window=window,
                      efermi=efermi)

        # Summ PDOS for the same atom types
        combine_by_atoms(input_path,
                         seedname,
                         output_path,
                         atoms,
                         wfcs,
                         case)

        # Plot summed PDOS
        for atom in atoms:
            for wfc_symbol, wfc_number in wfcs[(atom, atoms[atom][0])]:
                path = join(input_path,
                            seedname,
                            "summed-by-atom",
                            f"{atom}_wfc#{wfc_number}({wfc_symbol})")
                dos = np.loadtxt(path, skiprows=1).T
                if efermi == 0:
                    title = f"PDOS for {atom} ({wfc_symbol}) (0 is 0)"
                else:
                    title = f"PDOS for {atom} ({wfc_symbol}) (0 is Fermi energy)"
                if relative:
                    plot_projected_relative(wfc_symbol,
                                            dos,
                                            path,
                                            title,
                                            case,
                                            window=window,
                                            efermi=efermi,
                                            normalize=normalize)
                else:
                    plot_projected(wfc_symbol,
                                   dos,
                                   path,
                                   title,
                                   case,
                                   window=window,
                                   efermi=efermi)

        # Plot all individual plots
        if separate:
            for atom in atoms:
                for a_number in atoms[atom]:
                    for wfc_symbol, wfc_number in wfcs[(atom, a_number)]:
                        dos = np.loadtxt(join(input_path,
                                              f"{seedname}.pdos_atm#{a_number}({atom})_wfc#{wfc_number}({wfc_symbol})"),
                                         skiprows=1).T
                        if efermi == 0:
                            title = f"PDOS for {atom} #{a_number} ({wfc_symbol}) (0 is 0)"
                        else:
                            title = f"PDOS for {atom} #{a_number} ({wfc_symbol}) (0 is Fermi energy)"
                        if relative:
                            plot_projected_relative(wfc_symbol,
                                                    dos,
                                                    join(output_path, seedname,
                                                         "individual-plots",
                                                         f"pdos_atm#{a_number}({atom})_wfc#{wfc_number}({wfc_symbol})"),
                                                    title,
                                                    case,
                                                    window=window,
                                                    efermi=efermi,
                                                    normalize=normalize)
                        else:
                            plot_projected(wfc_symbol,
                                           dos,
                                           join(output_path, seedname,
                                                "individual-plots",
                                                f"pdos_atm#{a_number}({atom})_wfc#{wfc_number}({wfc_symbol})"),
                                           title,
                                           case,
                                           window=window,
                                           efermi=efermi)

        # Plot orbital contribution in the whole atom
        for atom in atoms:
            custom_dos = [np.loadtxt(join(input_path,
                                          f"{seedname}.pdos_tot"),
                                     skiprows=1).T[0]]
            wfc_symbols = []
            for wfc_symbol, wfc_number in wfcs[(atom, atoms[atom][0])]:
                wfc_symbols.append(f"#{wfc_number}, {wfc_symbol}")
                wfs_dos = np.loadtxt(join(input_path,
                                          seedname,
                                          "summed-by-atom",
                                          f"{atom}_wfc#{wfc_number}({wfc_symbol})"),
                                     skiprows=1).T
                if case == 0:
                    if len(custom_dos) == 1:
                        custom_dos.append(deepcopy(wfs_dos[1]))
                    else:
                        custom_dos[1] += wfs_dos[1]
                    custom_dos.append(wfs_dos[1])
                elif case in [1, 2]:
                    if len(custom_dos) == 1:
                        custom_dos.append(deepcopy(wfs_dos[1]))
                        custom_dos.append(deepcopy(wfs_dos[2]))
                    else:
                        custom_dos[1] += wfs_dos[1]
                        custom_dos[2] += wfs_dos[2]
                    custom_dos.append(wfs_dos[1])
                    custom_dos.append(wfs_dos[2])
            if efermi == 0:
                title = f"wfc contribution to {atom} (0 is 0)"
            else:
                title = f"wfc contribution to {atom} (0 is Fermi energy)"

            if relative:
                plot_projected_relative(atom,
                                        np.array(custom_dos),
                                        join(output_path, seedname,
                                             "summed-by-atom",
                                             "wfc-in-atoms", atom),
                                        title,
                                        case,
                                        window=window,
                                        efermi=efermi,
                                        normalize=normalize,
                                        labels=wfc_symbols,
                                        factor=len(wfc_symbols))
            else:
                plot_projected(atom,
                               np.array(custom_dos),
                               join(output_path, seedname,
                                    "summed-by-atom",
                                    "wfc-in-atoms", atom),
                               title,
                               case,
                               window=window,
                               efermi=efermi,
                               labels=wfc_symbols,
                               factor=len(wfc_symbols))

        # Plot orbital contribution in the whole atom (individual atoms)
        if separate:
            for atom in atoms:
                for a_number in atoms[atom]:
                    custom_dos = [np.loadtxt(join(input_path,
                                                  f"{seedname}.pdos_tot"),
                                             skiprows=1).T[0]]
                    wfc_symbols = []
                    for wfc_symbol, wfc_number in wfcs[(atom, atoms[atom][0])]:
                        wfc_symbols.append(f"#{wfc_number}, {wfc_symbol}")
                        wfs_dos = np.loadtxt(join(input_path,
                                                  f"{seedname}.pdos_atm#{a_number}({atom})_wfc#{wfc_number}({wfc_symbol})"),
                                             skiprows=1).T
                        if case == 0:
                            if len(custom_dos) == 1:
                                custom_dos.append(deepcopy(wfs_dos[1]))
                            else:
                                custom_dos[1] += wfs_dos[1]
                            custom_dos.append(wfs_dos[1])
                        elif case in [1, 2]:
                            if len(custom_dos) == 1:
                                custom_dos.append(deepcopy(wfs_dos[1]))
                                custom_dos.append(deepcopy(wfs_dos[2]))
                            else:
                                custom_dos[1] += wfs_dos[1]
                                custom_dos[2] += wfs_dos[2]
                            custom_dos.append(wfs_dos[1])
                            custom_dos.append(wfs_dos[2])
                    if efermi == 0:
                        title = f"wfc contribution to #{a_number} {atom} (0 is 0)"
                    else:
                        title = f"wfc contribution to #{a_number} {atom} (0 is Fermi energy)"

                    if relative:
                        plot_projected_relative(f"#{a_number} {atom}",
                                                np.array(custom_dos),
                                                join(output_path, seedname,
                                                     "individual-plots",
                                                     "wfc-in-atoms",
                                                     f"#{a_number}_{atom}"),
                                                title,
                                                case,
                                                window=window,
                                                efermi=efermi,
                                                normalize=normalize,
                                                labels=wfc_symbols,
                                                factor=len(wfc_symbols))
                    else:
                        plot_projected(f"#{a_number} {atom}",
                                       np.array(custom_dos),
                                       join(output_path, seedname,
                                            "individual-plots",
                                            "wfc-in-atoms",
                                            f"#{a_number} {atom}"),
                                       title,
                                       case,
                                       window=window,
                                       efermi=efermi,
                                       labels=wfc_symbols,
                                       factor=len(wfc_symbols))

        # Plot atom contribution to total
        custom_dos = np.loadtxt(join(input_path,
                                     f"{seedname}.pdos_tot"),
                                skiprows=1).T
        if case == 0:
            custom_dos = [custom_dos[0], custom_dos[2]]
        elif case == 1:
            custom_dos = [custom_dos[0], custom_dos[3], custom_dos[4]]
        elif case == 2:
            custom_dos = [custom_dos[0], custom_dos[2], custom_dos[3]]
        labels = []
        for atom in atoms:
            labels.append(atom)
            if case == 0:
                custom_dos.append(None)
            elif case in [1, 2]:
                custom_dos.append(None)
                custom_dos.append(None)
            for wfc_symbol, wfc_number in wfcs[(atom, atoms[atom][0])]:
                wfs_dos = np.loadtxt(join(input_path,
                                          seedname,
                                          "summed-by-atom",
                                          f"{atom}_wfc#{wfc_number}({wfc_symbol})"),
                                     skiprows=1).T
                if case == 0:
                    if custom_dos[-1] is None:
                        custom_dos[-1] = deepcopy(wfs_dos[1])
                    else:
                        custom_dos[-1] += wfs_dos[1]
                elif case in [1, 2]:
                    if custom_dos[-1] is None:
                        custom_dos[-2] = deepcopy(wfs_dos[1])
                        custom_dos[-1] = deepcopy(wfs_dos[2])
                    else:
                        custom_dos[-2] += wfs_dos[1]
                        custom_dos[-1] += wfs_dos[2]
            if efermi == 0:
                title = f"Atom contribution to total PDOS (0 is 0)"
            else:
                title = f"Atom contribution to total PDOS (0 is Fermi energy)"

            if relative:
                plot_projected_relative("Total PDOS",
                                        np.array(custom_dos),
                                        join(output_path, seedname,
                                             "summed-by-atom",
                                             "Atom-contributions"),
                                        title,
                                        case,
                                        window=window,
                                        efermi=efermi,
                                        normalize=normalize,
                                        labels=labels,
                                        factor=len(labels))
            else:
                plot_projected("Total PDOS",
                               np.array(custom_dos),
                               join(output_path, seedname,
                                    "summed-by-atom",
                                    "Atom-contributions"),
                               title,
                               case,
                               window=window,
                               efermi=efermi,
                               labels=labels,
                               factor=len(labels))

        # Plot atom contribution to total (individual atoms)
        if separate:
            custom_dos = np.loadtxt(join(input_path,
                                         f"{seedname}.pdos_tot"),
                                    skiprows=1).T
            if case == 0:
                custom_dos = [custom_dos[0], custom_dos[2]]
            elif case == 1:
                custom_dos = [custom_dos[0], custom_dos[3], custom_dos[4]]
            elif case == 2:
                custom_dos = [custom_dos[0], custom_dos[2], custom_dos[3]]
            labels = []
            for atom in atoms:
                for a_number in atoms[atom]:
                    labels.append(f"#{a_number} {atom}")
                    if case == 0:
                        custom_dos.append(None)
                    elif case in [1, 2]:
                        custom_dos.append(None)
                        custom_dos.append(None)
                    for wfc_symbol, wfc_number in wfcs[(atom, atoms[atom][0])]:
                        wfs_dos = np.loadtxt(join(input_path,
                                                  f"{seedname}.pdos_atm#{a_number}({atom})_wfc#{wfc_number}({wfc_symbol})"),
                                             skiprows=1).T
                        if case == 0:
                            if custom_dos[-1] is None:
                                custom_dos[-1] = deepcopy(wfs_dos[1])
                            else:
                                custom_dos[-1] += wfs_dos[1]
                        elif case in [1, 2]:
                            if custom_dos[-1] is None:
                                custom_dos[-2] = deepcopy(wfs_dos[1])
                                custom_dos[-1] = deepcopy(wfs_dos[2])
                            else:
                                custom_dos[-2] += wfs_dos[1]
                                custom_dos[-1] += wfs_dos[2]
                    if efermi == 0:
                        title = f"Atom contribution to total PDOS (0 is 0)"
                    else:
                        title = f"Atom contribution to total PDOS (0 is Fermi energy)"

                    if relative:
                        plot_projected_relative("Total PDOS",
                                                np.array(custom_dos),
                                                join(output_path, seedname,
                                                     "individual-plots",
                                                     "Atom-contributions"),
                                                title,
                                                case,
                                                window=window,
                                                efermi=efermi,
                                                normalize=normalize,
                                                labels=labels,
                                                factor=len(labels))
                    else:
                        plot_projected("Total PDOS",
                                       np.array(custom_dos),
                                       join(output_path, seedname,
                                            "individual-plots",
                                            "Atom-contributions"),
                                       title,
                                       case,
                                       window=window,
                                       efermi=efermi,
                                       labels=labels,
                                       factor=len(labels))

        print(f"{OK}" +
              f"Finish to work with {seedname} seedname, results are in " +
              f"  {abspath(join(output_path, seedname))}" +
              f"{RESET}")


def get_parser():
    parser = ArgumentParser(
        description="Script for visualisation of density of states.")
    parser.add_argument("-ip", "--input-path",
                        type=str,
                        required=True,
                        help="""
                        Relative or absulute path to the folder with dos files.
                        """
                        )
    parser.add_argument("-f", "--filpdos",
                        type=str,
                        default=None,
                        help="""
                        Prefix for output files containing PDOS(E). 
                        As specified in the QE projwfc.x input file.
                        """)
    parser.add_argument("-op", "--output-path",
                        type=str,
                        default='.',
                        help="""
                        Relative or absolute path to the folder
                        for saving outputs.
                        """
                        )
    parser.add_argument("-w", "--window",
                        type=float,
                        nargs=2,
                        default=None,
                        help="""
                        Energy window for the plots.  
                        By default whole range present in the files is plotted
                        """)
    parser.add_argument("-ef", "--efermi",
                        type=float,
                        default=0.,
                        help="""
                        Fermi energy, zero by default.
                        """)
    parser.add_argument("-s", "--separate",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to plot projected DOS for each atom 
                        of the same type separately.
                        """)
    parser.add_argument("-r", "--relative",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to use relative style.
                        """)
    parser.add_argument("-n", "--normalize",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to use normalize relative style.
                        """)

    return parser



r"""
Exchange model.

Write a tutorial with docstrings here.
"""

from copy import deepcopy
from math import sqrt, pi, cos, sin

import numpy as np

from rad_tools.exchange.template import ExchangeTemplate
from rad_tools.exchange.bond import Bond


class ExchangeModel:
    r"""
    Main entry point for the exchange model.

    Attributes
    ----------
    bonds : dict
        Dictionary of bonds.

        .. code-block:: python

            {(Atom_1, Atom_2, R): bond, ...}
    magnetic_atoms : dict
        Dictionary with keys : str - marks of atoms and value : 1 x 3 array
        - coordinate of the atom in Angstroms.
    noomagnetic_atoms : dist
        Dictionary with keys : str - marks of atoms and value : 1 x 3 array
        - coordinate of the atom in Angstroms. Only non-magnetic atoms.
    cell : 3 x 3 array of floats
    a : 3 x 1 array
    b : 3 x 1 array
    c : 3 x 1 array
    len_a : float
    len_b : float
    len_c : float
    unit_cell_volume : float
    b1 : 3 x 1 array
    b2 : 3 x 1 array
    b3 : 3 x 1 array
    cell_list : list
    number_spins_in_unit_cell : int

    space_dimensions : tuple of floats
    """

    def __init__(self) -> None:
        self._cell = np.identity(3, dtype=float)
        self.magnetic_atoms = {}
        self.nonmagnetic_atoms = {}
        self.bonds = {}

    def __iter__(self):
        return ExchangeModelIterator(self)

    def __contains__(self, item):
        return item in self.bonds

    def __getitem__(self, key) -> Bond:
        return self.bonds[key]

    @property
    def cell(self):
        r"""
        Matrix of lattice vectors in Angstrom.

        .. code-block:: python

            [[a_x  a_y  a_z],
             [b_x  b_y  b_z],
             [c_x  x_y  c_z]]

        Default cell is orthonormal.

        .. code-block:: python

            np.identity(3, dtype=float)
        """

        return self._cell

    @cell.setter
    def cell(self, new_cell):
        new_cell = np.array(new_cell)
        if new_cell.shape != (3, 3):
            raise ValueError("Cell matrix shape have to be equal (3, 3)")
        if np.linalg.det(new_cell) == 0:
            raise ValueError("Lattice vectors are complanar.")
        self._cell = new_cell

    @property
    def a(self):
        r"""
        First lattice vector :math:`\vec{a} = (a_x, a_y, a_z)`.
        """

        return self.cell[0]

    @a.setter
    def a(self, new_a):
        new_a = np.array(new_a)
        if new_a.shape != (3,):
            raise ValueError("new_a is not a 3 component vector")
        self._cell[0] = new_a

    @property
    def b(self):
        r"""
        Second lattice vector :math:`\vec{b} = (b_x, b_y, b_z)`.
        """

        return self.cell[1]

    @b.setter
    def b(self, new_b):
        new_b = np.array(new_b)
        if new_b.shape != (3,):
            raise ValueError("new_b is not a 3 component vector")
        self._cell[1] = new_b

    @property
    def c(self):
        r"""
        Third lattice vector :math:`\vec{b} = (b_x, b_y, b_z)`.
        """

        return self.cell[2]

    @c.setter
    def c(self, new_c):
        new_c = np.array(new_c)
        if new_c.shape != (3,):
            raise ValueError("new_c is not a 3 component vector")
        self._cell[2] = new_c

    @property
    def len_a(self):
        r"""
        Length of lattice vector :math:`\vec{a}`.
        """

        return sqrt(np.sum(self.a**2))

    @property
    def len_b(self):
        r"""
        Length of lattice vector :math:`\vec{b}`.
        """

        return sqrt(np.sum(self.b**2))

    @property
    def len_c(self):
        r"""
        Length of lattice vector :math:`\vec{c}`.
        """

        return sqrt(np.sum(self.c**2))

    @property
    def unit_cell_volume(self):
        r"""
        Volume of the unit cell.

        .. math::

            V = \delta_{\vec{A}}\cdot(\delta_{\vec{B}}\times\delta_{\vec{C}})
        """

        return np.dot(self.a, np.cross(self.b, self.c))

    @property
    def b1(self):
        r"""
        First reciprocal lattice vector

        .. math::

            \vec{b}_1 = \frac{2\pi}{V}\vec{b}\times\vec{c}

        where :math:`V = \vec{a}\cdot\vec{b}\times\vec{c}`
        """

        return 2 * pi / self.unit_cell_volume * np.cross(self.b, self.c)

    @property
    def b2(self):
        r"""
        Second reciprocal lattice vector

        .. math::

            \vec{b}_2 = \frac{2\pi}{V}\vec{c}\times\vec{a}

        where :math:`V = \vec{a}\cdot\vec{b}\times\vec{c}`
        """

        return 2 * pi / self.unit_cell_volume * np.cross(self.c, self.a)

    @property
    def b3(self):
        r"""
        Third reciprocal lattice vector

        .. math::

            \vec{b}_3 = \frac{2\pi}{V}\vec{a}\times\vec{b}

        where :math:`V = \vec{a}\cdot\vec{b}\times\vec{c}`
        """

        return 2 * pi / self.unit_cell_volume * np.cross(self.a, self.b)

    @property
    def cell_list(self):
        r"""
        List of cells from the model.

        Return the list of R for all cell that are present in the model.
        Not ordered.

        Returns
        -------
        cells : n x 3 array
            Array of n unit cells.
        """

        cells = set()
        for atom1, atom2, R in self.bonds:
            cells.add(R)
        return np.array(list(cells))

    @property
    def number_spins_in_unit_cell(self):
        r"""
        Number of spins (or magnetic atoms) in the unit cell.
        """

        return len(self.magnetic_atoms)

    @property
    def space_dimensions(self):
        r"""
        Model minimum and maximum coordiantes in real space.

        Takes into account only an atom which has at least one bond associated with it.

        Returns
        -------
        x_min : float
            Minimum x coordinate.
        y_min : float
            Minimum y coordinate.
        z_min : float
            Minimum z coordinate.
        x_max : float
            Maximum x coordinate.
        y_max : float
            Maximum y coordinate.
        z_max : float
            Maximum z coordinate.
        """

        x_max, y_max, z_max = None, None, None
        x_min, y_min, z_min = None, None, None
        for atom1, atom2, R in self.bonds:
            x1, y1, z1 = self.get_atom_coordinates(atom1)
            x2, y2, z2 = self.get_atom_coordinates(atom2, R)
            if x_max is None:
                x_min, y_min, z_min = x1, y1, z1
                x_max, y_max, z_max = x1, y1, z1
            x_max = max(x1, x2, x_max)
            y_max = max(y1, y2, y_max)
            z_max = max(z1, z2, z_max)
            x_min = min(x1, x2, x_min)
            y_min = min(y1, y2, y_min)
            z_min = min(z1, z2, z_min)
        return x_min, y_min, z_min, x_max, y_max, z_max

    def round(self, decimals=4):
        r"""Round exchange parameters.

        Parameters
        ----------
        decimals : int, default 4
            Number of decimals after the comma.
        """

        for atom1, atom2, R in self.bonds:
            self.bonds[(atom1, atom2, R)].round(decimals)

    def add_bond(self, bond: Bond, atom1, atom2, R):
        r"""
        Add one bond to the model.

        Parameters
        ----------
        bond : :py:class:`Bond`
            An instance of :py:class:`Bond` class with the information about
            exchange parameters.
        atom1 : str
            Name of atom1 in (0, 0, 0) unit cell.
        atom2 : str
            Name of atom2 in R unit cell.
        R : tuple of ints
            Vector of the unit cell for atom2. In the relative coordinates in real space.
        """

        self.bonds[(atom1, atom2, R)] = bond

    def remove_bond(self, atom1, atom2, R):
        r"""
        Remove one bond from the model.

        Parameters
        ----------
        atom1 : str
            Name of atom1 in (0, 0, 0) unit cell.
        atom2 : str
            Name of atom2 in R unit cell.
        R : tuple of ints
            Radius vector of the unit cell for atom2.
        """

        try:
            del self.bonds[(atom1, atom2, R)]
        except KeyError:
            pass

    def add_atom(self, name, a, b, c):
        r"""
        Add magnetic atom to the model.

        Parameters
        ----------
        name : str
            Mark for the atom. Note: if an atom with the same mark already
            exists in ``magnetic_atoms`` then it will be rewritten.
        a : int or float
            Relative coordinate of the atom along a lattice vector.
        b : int or float
            Relative coordinate of the atom along b lattice vector.
        c : int or float
            Relative coordinate of the atom along c lattice vector.
        """

        self.magnetic_atoms[name] = np.array([a, b, c])

    def remove_atom(self, name):
        r"""
        Remove magnetic atom from the model.

        Note: this method will remove atom from ``magnetic_atoms`` and all the
        bonds, which starts or ends in this atom.

        Parameters
        ----------
        name : str
            Mark for the atom.
        """

        bonds_for_removal = []
        if name in self.magnetic_atoms:
            del self.magnetic_atoms[name]
        for atom1, atom2, R in self.bonds:
            if atom1 == name or atom2 == name:
                bonds_for_removal.append((atom1, atom2, R))

        for bond in bonds_for_removal:
            del self.bonds[bond]

    def get_atom_coordinates(self, atom, R=(0, 0, 0)):
        r"""
        Getter for the atom coordinates.

        Parameters
        ----------
        atom : str
            Name of atom1 in R unit cell.
        R : 1 x 3 array, default (0, 0, 0)
            Radius vector of the unit cell for atom2.

        Returns
        -------
        coordinates : 1 x 3 array
            Coordinates of atom in the cell R in real space.
        """

        R = np.array(R)
        try:
            return np.matmul(R + self.magnetic_atoms[atom], self.cell)
        except KeyError:
            return np.matmul(R + self.nonmagnetic_atoms[atom], self.cell)

    def get_bond_centre_coordinates(self, atom1, atom2, R=(0, 0, 0)):
        r"""
        Getter for the middle point of the bond.

        Parameters
        ----------
        atom1 : str
            Name of atom1 in (0, 0, 0) unit cell.
        atom2 : str
            Name of atom2 in R unit cell.
        R : tuple of ints, default (0, 0, 0)
            Radius vector of the unit cell for atom2.

        Returns
        -------
        coordinates : 1 x 3 array
            Coordinates of the middle of a bond in real space.
        """

        atom1 = self.get_atom_coordinates(atom1)
        atom2 = self.get_atom_coordinates(atom2, R)

        return (atom1 + atom2) / 2

    def get_bond_vector(self, atom1, atom2, R=(0, 0, 0)):
        r"""
        Getter for vector between the atom1 and atom2.

        Parameters
        ----------
        atom1 : str
            Name of atom1 in (0, 0, 0) unit cell.
        atom2 : str
            Name of atom2 in R unit cell.
        R : tuple of ints, default (0, 0, 0)
            Radius vector of the unit cell for atom2.

        Returns
        -------
        v : 1 x 3 array
            Vector from atom1 to atom2 of the bond.
        """

        atom1 = self.get_atom_coordinates(atom1)
        atom2 = self.get_atom_coordinates(atom2, R)

        return atom2 - atom1

    def get_distance(self, atom1, atom2, R=(0, 0, 0)):
        r"""
        Getter for distance between the atom1 and atom2.

        Parameters
        ----------
        atom1 : str
            Name of atom1 in (0, 0, 0) unit cell.
        atom2 : str
            Name of atom2 in R unit cell.
        R : tuple of ints, default (0, 0, 0)
            Radius vector of the unit cell for atom2.

        Returns
        -------
        distance : floats
            Distance between atom1 and atom2.
        """

        return sqrt(np.sum(self.get_bond_vector(atom1, atom2, R)**2))

    def filter(self,
               max_distance=None,
               min_distance=None,
               template=None,
               R_vector=None):
        r"""
        Filter the exchange entries based on the given conditions.

        The result will be defined by logical conjugate of the conditions.
        Saying so the filtering will be performed for each given condition
        one by one.

        .. note::
            This method modifies the instance at which it is called.

        Parameters
        ----------
        max_distance : float or int, optional
            Distance for sorting, the condition is <<less or equal>>.
        min_distance : float or int, optional
            Distance for sorting, the condition is <<more or equal>>.
        template : list or :py:class:`.ExchangeTemplate`.
            List of pairs, which will remain in the model. ::

                [(atom1, atom2, R), ...]

        R_vector : tuple of ints or list of tuples of ints
            Tuple of 3 integers or list of tuples, specifying the R vectors,
            which will be kept after filtering.
        """

        if isinstance(template, ExchangeTemplate):
            template = template.get_list()
        if template is not None:
            template = set(template)
        if R_vector is not None:
            if type(R_vector) == tuple:
                R_vector = {R_vector}
            elif type(R_vector) == list:
                R_vector = set(R_vector)
            else:
                raise TypeError("Type is not supported, supported: list.")

        bonds_for_removal = set()
        for atom1, atom2, R in self.bonds:

            dis = self.get_distance(atom1, atom2, R)

            if max_distance is not None and dis > max_distance:
                bonds_for_removal.add((atom1, atom2, R))

            if min_distance is not None and dis < min_distance:
                bonds_for_removal.add((atom1, atom2, R))

            if R_vector is not None and R not in R_vector:
                bonds_for_removal.add((atom1, atom2, R))
            if template is not None and (atom1, atom2, R) not in template:
                bonds_for_removal.add((atom1, atom2, R))

        for atom1, atom2, R in bonds_for_removal:
            self.remove_bond(atom1, atom2, R)

    def filtered(self,
                 max_distance=None,
                 min_distance=None,
                 template=None,
                 R_vector=None):
        r"""
        Create filtered exchange model based on the given conditions.

        The result will be defined by logical conjugate of the conditions.
        Saying so the filtering will be performed for each given condition
        one by one.
        Note: this method is not modifying the instance at which it is called.
        It will create a new instance with sorted :py:attr:`bonds` and all the other
        attributes will be copied (through :py:func:`deepcopy`).

        .. note::
            This method is not modifying the instance at which it is called.
            It will create a new instance with merged ``bonds`` and all the
            other attributes will be copied (through deepcopy).

        Parameters
        ----------
        max_distance : float or int, optional
            Distance for sorting, the condition is <<less or equal>>.
        min_distance : float or int, optional
            Distance for sorting, the condition is <<more or equal>>.
        template : list or :py:class:`.ExchangeTemplate`
            List of pairs, which will remain in the model. ::

                [(atom1, atom2, R), ...]

        R_vector : tuple of ints or list of tuples of ints
            Tuple of 3 integers or list of tuples, specifying the R vectors,
            which will be kept after filtering.

        Returns
        -------
        filtered_model : :py:class:`.ExchangeModel`
            Exchange model after filtering.
        """

        filtered_model = deepcopy(self)
        filtered_model.filter(max_distance=max_distance,
                              min_distance=min_distance,
                              template=template,
                              R_vector=R_vector)
        return filtered_model

    def force_symmetry(self, template):
        r"""
        Force the model to have the symmetries of the template.

        Takes mean values of the parameters.
        For DMI interaction directions are kept the same,
        but the length of the DMI vector is scaled.

        This method modifies an instance on which it was called.

        Parameters
        ----------
        template : :py:class:`.ExchangeTemplate`
            Template.
        """

        if not isinstance(template, ExchangeTemplate):
            raise TypeError

        self.filter(template=template)

        for name in template.names:
            bonds = template.names[name]
            symm_matrix = np.zeros((3, 3), dtype=float)
            dmi_module = 0
            for bond in bonds:
                symm_matrix = (symm_matrix +
                               self.bonds[bond].symm_matrix)
                dmi_module = dmi_module + self.bonds[bond].dmi_module

            symm_matrix = symm_matrix / len(bonds)
            dmi_module = dmi_module / len(bonds)

            for bond in bonds:
                if self.bonds[bond].dmi_module != 0:
                    asymm_factor = dmi_module / self.bonds[bond].dmi_module
                else:
                    asymm_factor = 0
                self.bonds[bond].matrix = (symm_matrix +
                                           self.bonds[bond].asymm_matrix * asymm_factor)

    def forced_symmetry(self, template):
        r"""
        Force the model to have the symmetries of the template.

        Takes mean values of the parameters.
        Respect the direction of the DMI vectors.

        This method return a new instance.

        Parameters
        ----------
        template : :py:class:`.ExchangeTemplate`
            Template.

        Returns
        -------
        new_model : :py:class:`.ExchangeModel`
            Exchange model with forced symmetry.
        """

        new_model = deepcopy(self)
        new_model.force_symmetry(template=template)
        return new_model

    def summary_as_txt(self,
                       template: ExchangeTemplate,
                       decimals=4,
                       force_symmetry=False,
                       isotropic=False,
                       anisotropic=False,
                       out_matrix=False,
                       out_dmi=False):
        r"""
        Return exchange model based on the template file in .txt format.

        Parameters
        ----------
        template : :py:class:`.ExchangeTemplate`
            Template of the desired exchange model.
            (see :ref:`rad-make-template`)
        accuracy : int, default 4
            Accuracy for the exchange values
        force_symmetry: bool, default False
            Whenever to force the symmetry of the template exchange model.
            If ``False`` then each individual bond is written, otherwise
            exchange parameters of the template model are written.
        isotropic : bool, default False
            Whenever to output isotropic exchange.
        anisotropic : bool, default False
            Whenever to output anisotropic exchange.
        out_matrix : bool, default False
            Whenever to output whole matrix exchange.
        out_dmi : bool, default False
            Whenever to output DMI exchange.

        Returns
        -------
        summary : str
            Exchange information as a string.
        """

        if force_symmetry:
            self.force_symmetry(template=template)
        else:
            self.filter(template=template)
        self.round(decimals=decimals)
        summary = ""
        for name in template.names:
            scalar_written = False
            for atom1, atom2, R in template.names[name]:
                # Get values from the model
                bond = self.bonds[(atom1, atom2, R)]
                if not isinstance(bond, Bond):
                    raise TypeError
                iso = bond.iso
                aniso = bond.aniso
                dmi = bond.dmi
                abs_dmi = bond.dmi_module
                rel_dmi = abs_dmi/iso
                matrix = bond.matrix

                if not force_symmetry:

                    # Write the values
                    summary += (
                        f"{atom1:3} {atom2:3} " +
                        f"({R[0]:2.0f}, {R[1]:2.0f}, {R[2]:2.0f})\n")
                    if isotropic:
                        summary += f"    Isotropic: {iso:.{decimals}f}\n"
                    if anisotropic:
                        summary += (
                            f"    Anisotropic:\n" +
                            f"        {aniso[0][0]:{decimals+3}.{decimals}f}  " +
                            f"{aniso[0][1]:{decimals+3}.{decimals}f}  " +
                            f"{aniso[0][2]:{decimals+3}.{decimals}f}\n" +
                            f"        {aniso[1][0]:{decimals+3}.{decimals}f}  " +
                            f"{aniso[1][1]:{decimals+3}.{decimals}f}  " +
                            f"{aniso[1][2]:{decimals+3}.{decimals}f}\n" +
                            f"        {aniso[2][0]:{decimals+3}.{decimals}f}  " +
                            f"{aniso[2][1]:{decimals+3}.{decimals}f}  " +
                            f"{aniso[2][2]:{decimals+3}.{decimals}f}\n")
                    if out_matrix:
                        summary += (
                            f"    Matrix:\n" +
                            f"        {matrix[0][0]:{decimals+3}.{decimals}f}  " +
                            f"{matrix[0][1]:{decimals+3}.{decimals}f}  " +
                            f"{matrix[0][2]:{decimals+3}.{decimals}f}\n" +
                            f"        {matrix[1][0]:{decimals+3}.{decimals}f}  " +
                            f"{matrix[1][1]:{decimals+3}.{decimals}f}  " +
                            f"{matrix[1][2]:{decimals+3}.{decimals}f}\n" +
                            f"        {matrix[2][0]:{decimals+3}.{decimals}f}  " +
                            f"{matrix[2][1]:{decimals+3}.{decimals}f}  " +
                            f"{matrix[2][2]:{decimals+3}.{decimals}f}\n")
                    if out_dmi:
                        summary += (
                            f"    |DMI|: {abs_dmi:.{decimals}f}\n" +
                            f"    |DMI/J|: {rel_dmi:.{decimals}f}\n" +
                            f"    DMI: {dmi[0]:.{decimals}f} " +
                            f"{dmi[1]:.{decimals}f} " +
                            f"{dmi[2]:.{decimals}f}\n")
                    summary += "\n"
                else:
                    if not scalar_written:
                        scalar_written = True
                        summary += f"{name}\n"
                        if isotropic:
                            summary += f"    Isotropic: {iso:.{decimals}f}\n"
                        if anisotropic:
                            summary += (
                                f"    Anisotropic:\n" +
                                f"        {aniso[0][0]:{decimals+3}.{decimals}f}  " +
                                f"{aniso[0][1]:{decimals+3}.{decimals}f}  " +
                                f"{aniso[0][2]:{decimals+3}.{decimals}f}\n" +
                                f"        {aniso[1][0]:{decimals+3}.{decimals}f}  " +
                                f"{aniso[1][1]:{decimals+3}.{decimals}f}  " +
                                f"{aniso[1][2]:{decimals+3}.{decimals}f}\n" +
                                f"        {aniso[2][0]:{decimals+3}.{decimals}f}  " +
                                f"{aniso[2][1]:{decimals+3}.{decimals}f}  " +
                                f"{aniso[2][2]:{decimals+3}.{decimals}f}\n")
                        if out_matrix:
                            summary += (
                                f"    Matrix:\n" +
                                f"        {matrix[0][0]:{decimals+3}.{decimals}f}  " +
                                f"{matrix[0][1]:{decimals+3}.{decimals}f}  " +
                                f"{matrix[0][2]:{decimals+3}.{decimals}f}\n" +
                                f"        {matrix[1][0]:{decimals+3}.{decimals}f}  " +
                                f"{matrix[1][1]:{decimals+3}.{decimals}f}  " +
                                f"{matrix[1][2]:{decimals+3}.{decimals}f}\n" +
                                f"        {matrix[2][0]:{decimals+3}.{decimals}f}  " +
                                f"{matrix[2][1]:{decimals+3}.{decimals}f}  " +
                                f"{matrix[2][2]:{decimals+3}.{decimals}f}\n")
                        if out_dmi:
                            summary += (
                                f"    |DMI|: {abs_dmi:.{decimals}f}\n" +
                                f"    |DMI/J|: {rel_dmi:.{decimals}f}\n")
                    if out_dmi:
                        summary += (
                            f"    DMI: {dmi[0]:{decimals+3}.{decimals}f} " +
                            f"{dmi[1]:{decimals+3}.{decimals}f} " +
                            f"{dmi[2]:{decimals+3}.{decimals}f} " +
                            f"({atom1:3} {atom2:3} " +
                            f"{R[0]:2.0f} {R[1]:2.0f} {R[2]:2.0f})\n")
            if force_symmetry:
                summary += "\n"

        return summary

    def ferromagnetic_energy(self, theta=0, phi=0):
        r"""
        Compute energy of the model assuming ferromagnetic state.

        With Hamiltonian of the form:

        .. math::

            \hat{H} = - \sum_{i,j} \vec{S}_i J_{ij} \vec{S}_j

        where spin vectors are normaized to 1 and :math:`J_{ij}` is the 
        exchange matrix.

        Parameters
        ----------
        theta : float, default 0
            Angle between z axis an direction of the magnetization.
            :math:`0 < \theta < 180`
        phi : float, default 0
            angle between x axis an projection of direction of the 
            magnetization on xy plane.
            :math:`0 < \phi < 360`

        Returns
        -------
        energy : float
            Energy of ferromagnetic model with magnetisation direction defined 
            by ``theta`` and ``phi``. In the units of J values.
        """

        theta = theta / 180 * pi
        phi = phi / 180 * pi
        energy = np.zeros((3, 3), dtype=float)
        for bond in self:
            energy -= self.bonds[bond].matrix
        spin_vector = np.array([cos(phi) * sin(theta),
                                sin(phi) * sin(theta),
                                cos(theta)])
        energy = np.matmul(np.matmul(spin_vector, energy), spin_vector)
        return energy


class ExchangeModelIterator:

    def __init__(self, exchange_model: ExchangeModel) -> None:
        self._bonds = list(exchange_model.bonds)
        self._index = 0

    def __next__(self):
        if self._index < len(self._bonds):
            result = self._bonds[self._index]
            self._index += 1
            return result
        raise StopIteration

    def __iter__(self):
        return self



r"""
Exchange template.
"""


class ExchangeTemplate:
    r"""
    Store a template for sorting the exchange from TB2J file.

    In addition stores the technical details for plotting, 
    orbital decompozition, etc.

    Attributes
    ----------
    names : dict
        Dictionary of neighbours from the template file.

        .. code-block:: python

            {name : [(atom1, atom2, R), ...], ...}

    latex_names : dict
        The dictionary of Letex version of names from `names`.

        .. code-block:: python

            {name : latex_name, ...}
    """

    def __init__(self) -> None:

        self.names = {}
        self.latex_names = {}

    def get_list(self):
        r"""
        Translate bonds present in the template into the list.

        Returns
        -------
        plained_template : list
            List with the bond specifications:

            .. code-block:: python

                [(atom_1, atom_2, R), ...]
        """
        plained_template = []
        for name in self.names:
            for atom1, atom2, R in self.names[name]:
                plained_template.append((atom1, atom2, R))
        return plained_template



r"""
Collection of small routines and constants, 
which may be used across the whole package.
"""

import numpy as np
from math import asin, sqrt, pi

# Teminal colors
BLACK = "\u001b[30m"
"""
ANSI escape code for black color of the text.
"""
RED = "\u001b[31m"
"""
ANSI escape code for red color of the text.
"""
GREEN = "\u001b[32m"
"""
ANSI escape code for green color of the text.
"""
YELLOW = "\u001b[33m"
"""
ANSI escape code for yellow color of the text.
"""
BLUE = "\u001b[34m"
"""
ANSI escape code for blue color of the text.
"""
MAGENTA = "\u001b[35m"
"""
ANSI escape code for magenta color of the text.
"""
CYAN = "\u001b[36m"
"""
ANSI escape code for cyan color of the text.
"""
WHITE = "\u001b[37m"
"""
ANSI escape code for white color of the text.
"""
RESET = "\u001b[0m"
"""
ANSI escape code for resetting color to default.
"""
WARNING = YELLOW
"""
ANSI escape code for warnings.
"""
OK = GREEN
"""
ANSI escape code for ok messages.
"""
ERROR = RED
"""
ANSI escape code for errors.
"""


def get_256_colours(n):
    r"""
    ANSI escape codes for terminal color with 256-colours support
    (see: :ANSI:`wiki <>`).

    Parameters
    ----------
    n : int
        Integer from 0 to 255 to be mapped to colours.

    Returns
    -------
    str
        string with the ANSI escape code.
    """

    if type(n) != int or not 0 <= n <= 255:
        raise ValueError(f'Integer n have to be in range 0 <= n <= 255. '
                         f'You provided n = {n}, type<n> = {type(n)}')
    return f'\033[38:5:{n}m'


def atom_mark_to_latex(mark):
    r"""
    Latexifier for atom marks. 

    Cr12 -> Cr\ :sub:`12`\.

    Parameters
    ----------
    mark : str
        Mark of atom.

    Returns
    -------
    new_mark : str
        Latex version of the mark.
    """
    numbers = '0123456789'
    new_mark = '$'
    insert_underline = False
    for symbol in mark:
        if symbol in numbers and not insert_underline:
            insert_underline = True
            new_mark += '_{'
        new_mark += symbol
    new_mark += '}$'
    return new_mark


def rot_angle(x, y, dummy=False):
    r"""
    Rotational ange from 2D vector.

    Mathematically positive => counterclockwise.
    From [0 to 360)

    Parameters
    ----------
    x : float or int
        x coordinate of a vector.
    y : float or int
        y coordinate of a vector.
    """
    # rot_cos = x / (x ** 2 + y ** 2) ** 0.5
    # rot_angle = m.acos(rot_cos) / m.pi * 180
    try:
        sin = abs(y) / sqrt(x**2 + y**2)
    except ZeroDivisionError:
        raise ValueError('Angle is ill defined (x = y = 0).')
    if x > 0:
        if y > 0:
            return asin(sin) / pi * 180
        elif y == 0:
            return 0
        elif y < 0:
            if not dummy:
                return -asin(sin) / pi * 180
            return 360 - asin(sin) / pi * 180
    elif x == 0:
        if y > 0:
            return 90
        elif y == 0:
            raise ValueError('Angle is ill defined (x = y = 0).')
        elif y < 0:
            if not dummy:
                return 90
            return 270
    elif x < 0:
        if y > 0:
            if not dummy:
                return -asin(sin) / pi * 180
            return 180 - asin(sin) / pi * 180
        elif y == 0:
            if not dummy:
                return 0
            return 180
        elif y < 0:
            if not dummy:
                return asin(sin) / pi * 180
            return 180 + asin(sin) / pi * 180


def strip_digits(line: str):
    r"""
    Remove all digits from the string

    Parameters
    ----------
    line : str
        Input string.

    Returns
    -------
    new_line : str
        ``line`` without digits
    """

    new_line = ""
    numbers = "1234567890"
    for char in line:
        if char not in numbers:
            new_line += char
    return new_line


def two_points_distance(point1, point2):
    r"""
    Compute distance between two points.

    Parameters
    ----------
    point1 : array
        Coordinates of the first point. 

        .. code-block:: python

            [x1, y1, z1]

    point2 : array
        Coordinates of the second point. 

        .. code-block:: python

            [x2, y2, z2]

    Returns
    -------
    distance : float
        Distance between two points.
    """

    point1 = np.array(point1)
    point2 = np.array(point2)
    return sqrt(np.sum((point1 - point2)**2))


def search_on_atoms(centre, atoms):
    r"""
    Search the closest atom to the given centre position.

    Parameters
    ----------
    centre : array
        xyz coordinates of the centre.
    atoms : list
        List of atom names and coordinates of the form: ::

            [(name, [x, y, z]), ...]

    Returns
    -------
    min_span : float
        Distance from the centre to the closest atom.
    name : str
        Name of the closest atom.
    """

    min_span = 10000
    name = "None"
    for atom, a_coord in atoms:
        if sqrt(np.sum((centre[1] - a_coord)**2)) < min_span:
            min_span = sqrt(np.sum((centre[1] - a_coord)**2))
            name = atom
    return min_span, name


def search_between_atoms(centre, atoms):
    r"""
    Search the closest bond centre to the given centre position.

    Parameters
    ----------
    centre : array
        xyz coordinates of the centre.
    atoms : list
        List of atom names and coordinates of the form: ::

            [(name, [x, y, z]), ...]

    Returns
    -------
    min_span : float
        Distance from the centre to the bond`s centre.
    name : str
        Name of the closest bond`s centre (atom1-atom2).
    """

    pairs = []
    for i, atom in enumerate(atoms):
        for j in range(i+1, len(atoms)):
            pair = f"{atom[0]}-{atoms[j][0]}"
            p_coord = (atom[1]+atoms[j][1])/2
            pairs.append([pair, p_coord])
    return search_on_atoms(centre, pairs)


def absolute_to_relative(cell, x, y, z):
    r"""
    Compute relative coordinates with respect to the unit cell.

    Parameters
    ----------
    cell : 3 x 3 array
        Lattice vectors.
    x : float
        x coordinate.
    y : float
        y coordinate.
    z : float
        z coordinate.

    Returns
    -------
    relative : 1 x 3 array
        Relative coordinate.
    """

    a = np.array(cell[0], dtype=float)
    b = np.array(cell[1], dtype=float)
    c = np.array(cell[2], dtype=float)
    v = np.array([x, y, z], dtype=float)
    if (v == np.zeros(3)).all():
        return np.zeros(3)
    B = np.array([np.dot(a, v), np.dot(b, v), np.dot(c, v)])
    A = np.array([[np.dot(a, a), np.dot(a, b), np.dot(a, c)],
                  [np.dot(b, a), np.dot(b, b), np.dot(b, c)],
                  [np.dot(c, a), np.dot(c, b), np.dot(c, c)]])
    relative = np.linalg.solve(A, B)
    return relative



r"""
Module for constracting the k-points paths and high-symmetry k-points [1]_. 

Examples
--------

.. code-block:: python

    from rad_tools.kpoints import HighSymmetryPoints

    hexagonal = HighSymmetryPoints().hex()

References
----------
.. [1] 
    Setyawan, W. and Curtarolo, S., 2010. 
    High-throughput electronic band structure calculations: Challenges and tools. 
    Computational materials science, 49(2), pp.299-312.

    :DOI:`10.1016/j.commatsci.2010.05.010`
"""

from rad_tools.kpoints.high_symmetry_point import HighSymmetryPoints
from rad_tools.kpoints.kpoints import KPoints



#! /usr/local/bin/python3

from argparse import ArgumentParser
from os.path import split, join

import numpy as np

from rad_tools.routines import strip_digits, WARNING, RESET, \
    search_on_atoms, search_between_atoms


def identify(filename, span, output_path, out_name, nocolor=False):
    separation_tolerance = 10E-5

    # Read atoms and centres
    atom_counter = {}
    with open(filename, "r") as file:
        n = int(file.readline())
        file_stats = file.readline()
        centres = []
        atoms = []
        for _ in range(0, n):
            line = file.readline().split()
            if line[0] == "X":
                centres.append([f"X",
                                np.array(list(map(float, line[1:])))])
            else:
                if line[0] in atom_counter:
                    atom_counter[line[0]] += 1
                else:
                    atom_counter[line[0]] = 1
                atoms.append([f"{line[0]}{atom_counter[line[0]]}",
                              np.array(list(map(float, line[1:])))])

    # Identify centres localization
    for centre in centres:
        min_span_atom, atom = search_on_atoms(
            centre, atoms)
        min_span_pair, pair = search_between_atoms(
            centre, atoms)
        if min_span_atom > span and min_span_pair > span:
            if not nocolor:
                print(f"{WARNING}", end="")
            print(f"Centre {centre} unindentified, " +
                  "try to increase span")
            if not nocolor:
                print(f"{RESET}", end="")
            print(f"    span limit = {span}\n" +
                  f"    minimum distance to the atom = {min_span_atom:.8f} " +
                  f"({atom})\n"
                  f"    minimum distance to the bond`s centre = {min_span_pair:.8f} " +
                  f"({pair})\n")
            if abs(min_span_atom - min_span_pair) < separation_tolerance:
                centre[0] = f"None, closest: {atom}"
            elif min_span_atom < min_span_pair:
                centre[0] = f"None, closest: {atom}"
            else:
                centre[0] = f"None, closest: {pair}"
        elif abs(min_span_atom - min_span_pair) < separation_tolerance:
            centre[0] = f"{atom} or {pair}"
        elif min_span_atom < min_span_pair:
            centre[0] = atom
        else:
            centre[0] = pair

    # Write the output
    with open(join(output_path, out_name), "w") as file:
        file.write(f"{len(atoms) + len(centres):6.0f}\n")
        file.write(f"{file_stats}")
        for centre, coordinate in centres:
            file.write(f"X      " +
                       f"{coordinate[0]:14.8f}   " +
                       f"{coordinate[1]:14.8f}   " +
                       f"{coordinate[2]:14.8f}" +
                       f"   ->   {centre}\n")
        for atom, coordinate in atoms:
            file.write(f"{strip_digits(atom):2}     " +
                       f"{coordinate[0]:14.8f}   " +
                       f"{coordinate[1]:14.8f}   " +
                       f"{coordinate[2]:14.8f}" +
                       f"   ->   {atom}\n")


if __name__ == "__main__":
    parser = ArgumentParser(
        description=("Identify wannier centres with respect to the atom " +
                     "or to the point between the atom`s pair"))
    parser.add_argument("filename",
                        type=str,
                        help="""
                        Rellative or absolute path to the _centres.xyz file
                        """)
    parser.add_argument("-s", "--span",
                        type=float,
                        default=0.1,
                        help="""
                        Distance tolerance between centre and atom. (in Angstrom)
                        """)
    parser.add_argument("-op", "--output-path",
                        type=str,
                        default=None,
                        help="""
                        Relative or absolute path to the folder
                        for saving outputs.
                        """
                        )
    parser.add_argument("-on", "--output-name",
                        type=str,
                        default=None,
                        help="""
                        Seedname for the output files.
                        """
                        )
    parser.add_argument("-nc", "--no-colour",
                        action="store_true",
                        default=False,
                        help="""
                        Turn off coloured output.
                        """
                        )
    args = parser.parse_args()
    print(args)

    head, tail = split(args.filename)
    if args.output_path is None:
        args.output_path = head
    if args.output_name is None:
        args.output_name = tail + "_identified"

    identify(args.filename, args.span,
             output_path=args.output_path, out_name=args.output_name,
             nocolor=args.no_colour)



#! /usr/local/bin/python3

from argparse import ArgumentParser, RawTextHelpFormatter
from subprocess import run


def J_values(FM, AFM1, AFM2, AFM3):

    J2 = -(FM - AFM1)*13605/8/8
    J3 = -(FM - AFM2)*13605/2/2/8 - J2
    J1 = -(AFM2 - AFM3)*13605/8/2/8 + J3

    return J1, J2, J3


def compute(filename):
    FM = {}
    AFM1 = {}
    AFM2 = {}
    AFM3 = {}
    file = open(filename, "r")
    for line in file:
        mode = int(line.split("mode")[0])
        conf = line.split("/")[1]
        negpos = line.split("crsbr.scf")[1].split(".out")[0]
        energy = float(line.split("=")[1].split()[0])
        if "neg" in negpos:
            negpos = -int(negpos.split("neg")[0])
        elif "pos" in negpos:
            negpos = int(negpos.split("pos")[0])

        if mode not in FM:
            FM[mode] = {}
        if mode not in AFM1:
            AFM1[mode] = {}
        if mode not in AFM2:
            AFM2[mode] = {}
        if mode not in AFM3:
            AFM3[mode] = {}

        if conf == "FM":
            FM[mode][negpos] = energy
        elif conf == "AF1":
            AFM1[mode][negpos] = energy
        elif conf == "AF2":
            AFM2[mode][negpos] = energy
        elif conf == "AF3":
            AFM3[mode][negpos] = energy

    modes_list = [key for key in FM]
    modes_list.sort()
    print("MODES:\n")
    print(f"    {modes_list}\n\n")
    print("DISPLACEMENTS:\n")
    for mode in modes_list:
        dis_list = [key for key in FM[mode]]
        dis_list.sort()
        print(f"    mode {mode}: {dis_list}")
    print("\n")

    for mode in modes_list:
        dis_list = [key for key in FM[mode]]
        dis_list.sort()
        for dis in dis_list:
            ISOK = True
            try:
                fm = FM[mode][dis]
            except KeyError:
                print(
                    f"mode {mode}, displacement {dis/100:.2f} FM : problem with scf")
                ISOK = False
            try:
                afm1 = AFM1[mode][dis]
            except KeyError:
                print(
                    f"mode {mode}, displacement {dis/100:.2f} AFM1 : problem with scf")
                ISOK = False
            try:
                afm2 = AFM2[mode][dis]
            except KeyError:
                print(
                    f"mode {mode}, displacement {dis/100:.2f} AFM2 : problem with scf")
                ISOK = False
            try:
                afm3 = AFM3[mode][dis]
            except KeyError:
                print(
                    f"mode {mode}, displacement {dis/100:.2f} AFM3 : problem with scf")
                ISOK = False
            if ISOK:
                J1, J2, J3 = J_values(fm, afm1, afm2, afm3)
                print(f"Mode {mode}, displacement {dis/100:.2f}")
                print(f"    J1 = {J1:.6f} J2 = {J2:.6f} J3 = {J3:.6f}")


if __name__ == "__main__":
    parser = ArgumentParser(formatter_class=RawTextHelpFormatter)
    parser.add_argument("filename", type=str,
                        help="""File with grepped lines.
Example grep:

grep "!" *mode/*F*/{neg,pos}/crsbr.scf*out > summary.txt
    """)
    args = parser.parse_args()

    compute(args.filename)



from argparse import ArgumentParser
from os.path import join, abspath
from os import makedirs
from datetime import datetime
from calendar import month_name


from rad_tools.io.internal import read_template
from rad_tools.io import read_tb2j_model
from rad_tools.routines import OK, RESET


def manager(input_filename,
            template_file,
            output_path=".",
            output_name=None,
            accuracy=4,
            force_symmetry=False,
            isotropic=False,
            anisotropic=False,
            matrix=False,
            dmi=False,
            all=False):
    r"""
    Main function call of the tb2j-extractor.py script.

    Full documentation on the behaviour is available 
    :ref:`here <tb2j-extractor>`. Parameters of the function directly 
    corresponds to the arguments of the script.

    If you want to have the behaviour of the tb2j-extractor.py script 
    but in a format of a fuction call use this function.
    """

    try:
        makedirs(output_path)
    except FileExistsError:
        pass

    cd = datetime.now()

    if all:
        isotropic = True
        anisotropic = True
        matrix = True
        dmi = True

    model = read_tb2j_model(input_filename, quiet=True)
    template = read_template(template_file)
    summary_txt = model.summary_as_txt(template=template,
                                       decimals=accuracy,
                                       force_symmetry=force_symmetry,
                                       isotropic=isotropic,
                                       anisotropic=anisotropic,
                                       out_matrix=matrix,
                                       out_dmi=dmi)

    if output_name is not None:
        with open(join(output_path, output_name + ".txt"), "w") as out_file:
            out_file.write(
                f"Exchange values are extracted from: {input_filename}\n" +
                f"on {cd.day} {month_name[cd.month]} {cd.year}" +
                f" at {cd.hour}:{cd.minute}:{cd.second} by rad-tools\n\n")
            out_file.write(summary_txt)
        print(f"{OK}Extracted exchange info is in " +
              f"{abspath(join(output_path, output_name + '.txt'))}{RESET}")
    else:
        print(f"{summary_txt}")


def get_parser():
    parser = ArgumentParser(
        description="Script for extracting of template-based model from TB2J results.")

    parser.add_argument("-if", "--input-filename",
                        type=str,
                        required=True,
                        help="""
                        Relative or absulute path to the exchange.out file,
                        including the name and extention of the file itself.
                        """)
    parser.add_argument("-tf", "--template-file",
                        type=str,
                        required=True,
                        help="""
                        Relative or absolute path to the template file,
                        including the name and extention of the file.
                        """)
    parser.add_argument("-op", "--output-path",
                        type=str,
                        default=".",
                        help="""
                        Relative or absolute path to the folder
                        for saving outputs.
                        """
                        )
    parser.add_argument("-on", "--output-name",
                        type=str,
                        default=None,
                        help="""
                        Seedname for the output files.
                        """
                        )
    parser.add_argument("-acc", "--accuracy",
                        type=int,
                        default=4,
                        help="""
                        Accuracy for the exchange values.
                        """
                        )
    parser.add_argument("-fs", "--force-symmetry",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to force the symmetry of the
                        template on the model.
                        """)
    parser.add_argument("-i", "--isotropic",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to output isotropic exchange.
                        """)
    parser.add_argument("-a", "--anisotropic",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to output anisotropic exchange.
                        """)
    parser.add_argument("-m", "--matrix",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to output whole matrix exchange.
                        """)
    parser.add_argument("-dmi",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to output DMI exchange.
                        """)
    parser.add_argument("-all",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to all types of exchange.
                        """)

    return parser



#! /usr/local/bin/python3

from rad_tools.score.rad_dos_plotter_core import get_parser, manager

if __name__ == '__main__':
    parser = get_parser()
    args = parser.parse_args()
    manager(**vars(args))



from argparse import ArgumentParser
from os.path import join, abspath
from os import makedirs
from math import sqrt

from matplotlib import pyplot as plt
import numpy as np

from rad_tools.io.tb2j import read_exchange_model
from rad_tools.io.internal import read_template
from rad_tools.routines import atom_mark_to_latex, rot_angle, OK, RESET


def manager(input_filename,
            output_path='.',
            output_name='exchange',
            what_to_plot='iso',
            draw_cells=False,
            min_distance=None,
            max_distance=None,
            distance=None,
            template_file=None,
            R_vector=None,
            scale_atoms=1,
            scale_data=1,
            title=None,
            force_symmetry=False):

    if force_symmetry and template_file is None:
        raise ValueError("--force-symmetry option requires a template file.")

    if distance is not None:
        min_distance = distance
        max_distance = distance

    try:
        makedirs(output_path)
    except FileExistsError:
        pass

    if R_vector is not None:
        R_vector = np.array(R_vector[:len(R_vector) // 3 * 3],
                            dtype=int).reshape((len(R_vector)//3, 3))
        R_vector = list(map(tuple, R_vector.tolist()))

    model = read_exchange_model(input_filename)
    if template_file is not None:
        template = read_template(template_file)
    else:
        template = None
    model.filter(min_distance=min_distance,
                 max_distance=max_distance,
                 R_vector=R_vector,
                 template=template)
    if force_symmetry:
        model.force_symmetry(template=template)

    dummy = True
    ha = 'center'
    x_min, y_min, z_min, x_max, y_max, z_max = model.space_dimensions
    X = max(abs(x_min), abs(x_max))
    Y = max(abs(y_min), abs(y_max))
    if X == 0 and Y == 0:
        X = Y = 1
    linewidth = 1
    fontsize = 11
    if what_to_plot == "all":
        wtps = ['iso', 'distance', 'dmi']
    else:
        wtps = [what_to_plot]
    for wtp in wtps:

        if X < Y:
            fig = plt.figure(figsize=(6.4*X/Y, 4.8))
        else:
            fig = plt.figure(figsize=(6.4, 4.8*Y/X))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.set_aspect("equal")
        ax.set_xlabel('x, Angstroms')
        ax.set_ylabel('y, Angstroms')

        for atom1, atom2, R in model:
            bond = model[(atom1, atom2, R)]
            dis = model.get_distance(atom1, atom2, R)
            x1, y1, z1 = model.get_atom_coordinates(atom1)
            x2, y2, z2 = model.get_atom_coordinates(atom2, R)
            xm = (x1 + x2) / 2
            ym = (y1 + y2) / 2
            zm = (z1 + z2) / 2

            ax.scatter(x1, y1, s=100 * fig.dpi/72., c='white')
            ax.scatter(x2, y2, s=100 * fig.dpi/72., c='white')

            ax.text(x1, y1, atom_mark_to_latex(atom1),
                    va='center', ha='center',
                    fontsize=1.5 * fontsize * scale_atoms)
            ax.text(x2, y2, atom_mark_to_latex(atom2),
                    va='center', ha='center',
                    fontsize=1.5 * fontsize * scale_atoms)
            if wtp == 'iso':
                ax.text(xm, ym, str(round(bond.iso, 4)),
                        va='bottom', ha=ha,
                        rotation_mode='anchor',
                        rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                        fontsize=fontsize * scale_data)
            elif wtp == 'distance':
                ax.text(xm, ym, str(round(dis, 4)),
                        va='bottom', ha=ha,
                        rotation_mode='anchor',
                        rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                        fontsize=fontsize * scale_data)
            elif wtp == 'dmi':
                ax.text(xm, ym, str(round(sqrt(np.sum(bond.dmi**2)), 4)),
                        va='bottom', ha=ha,
                        rotation_mode='anchor',
                        rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                        fontsize=fontsize * scale_data)

        xlims = (ax.get_xlim()[0] - 0.5, ax.get_xlim()[1] + 0.5)
        ylims = (ax.get_ylim()[0] - 0.5, ax.get_ylim()[1] + 0.5)

        if draw_cells:
            cells = model.cell_list
            a_x, a_y, a_z = tuple(model.cell[0])
            b_x, b_y, b_z = tuple(model.cell[1])
            Rx_min = 0
            Rx_max = 0
            Ry_min = 0
            Ry_max = 0
            Rz_min = 0
            Rz_max = 0
            for Rx, Ry, Rz in cells:
                Rx_min = min(Rx, Rx_min)
                Rx_max = max(Rx, Rx_max)
                Ry_min = min(Ry, Ry_min)
                Ry_max = max(Ry, Ry_max)
                Rz_min = min(Rz, Rz_min)
                Rz_max = max(Rz, Rz_max)

            for i in range(Rx_min, Rx_max + 2):
                for j in range(Ry_min, Ry_max + 2):
                    for k in range(Rz_min, Rz_max + 2):
                        ax.plot([Rx_min * a_x + j * b_x,
                                (Rx_max + 1) * a_x + j * b_x],
                                [Rx_min * a_y + j * b_y,
                                (Rx_max + 1) * a_y + j * b_y],
                                linewidth=linewidth, color="#DFDFDF")
                        ax.plot([i * a_x + Ry_min * b_x,
                                i * a_x + (Ry_max + 1) * b_x],
                                [i * a_y + Ry_min * b_y,
                                i * a_y + (Ry_max + 1) * b_y],
                                linewidth=linewidth, color="#DFDFDF")
            ax.plot(np.array([0, a_x, a_x + b_x, b_x, 0]),
                    np.array([0, a_y, a_y + b_y, b_y, 0]),
                    linewidth=linewidth, color="#CA7371")

        ax.set_xlim(*xlims)
        ax.set_ylim(*ylims)

        if title is not None:
            ax.set_title(title, fontsize=1.5 * fontsize)

        png_path = join(output_path, f'{output_name}.{wtp}.png')
        plt.savefig(png_path, dpi=400, bbox_inches="tight")
        print(f'{OK}2D plot with {wtp} is in {abspath(png_path)}{RESET}')


def get_parser():
    parser = ArgumentParser(
        description="Script for visualisation of TB2J results")

    parser.add_argument("-if", "--input-filename",
                        type=str,
                        required=True,
                        help="""
                        Relative or absulute path to the *exchange.out* file,
                        including the name and extention of the file itself.
                        """
                        )
    parser.add_argument("-op", "--output-path",
                        type=str,
                        default='.',
                        help="""
                        Relative or absolute path to the folder for saving outputs.
                        """
                        )
    parser.add_argument("-on", "--output-name",
                        type=str,
                        default='exchange',
                        help="""
                        Seedname for the output files.
                        """
                        )
    parser.add_argument("-wtp", "--what-to-plot",
                        type=str,
                        choices=['all', 'iso', 'distance', 'dmi'],
                        default='all',
                        help="""
                        Type of data for display.
                        """
                        )
    parser.add_argument("-dc", "--draw-cells",
                        action="store_true",
                        default=False,
                        help="""
                        Whenever to draw the cells.
                        """
                        )
    parser.add_argument("-R", "--R-vector",
                        type=int,
                        nargs="*",
                        default=None,
                        help="""
                        R vectors for filtering the model.
                        """
                        )
    parser.add_argument("-maxd", "--max-distance",
                        type=float,
                        default=None,
                        help="""
                        (<=) Maximum distance.
                        """
                        )
    parser.add_argument("-mind", "--min-distance",
                        type=float,
                        default=None,
                        help="""
                        (>=) Minimum distance.
                        """
                        )
    parser.add_argument("-d", "--distance",
                        type=float,
                        default=None,
                        help="""
                        (=) Exact distance.
                        """
                        )
    parser.add_argument("-tf", "--template-file",
                        type=str,
                        default=None,
                        help="""
                        Relative or absolute path to the template file,
                        including the name and extention of the file.
                        """)
    parser.add_argument("-sa", "--scale-atoms",
                        default=1,
                        type=float,
                        help="""
                        Scale for the size of atom marks.
                        """)
    parser.add_argument("-sd", "--scale-data",
                        default=1,
                        type=float,
                        help="""
                        Scale for the size of data text.
                        """)
    parser.add_argument("-t", "--title",
                        default=None,
                        type=str,
                        help="""
                        Title for the plots.
                        """)
    parser.add_argument("-fs", "--force-symmetry",
                        default=False,
                        action="store_true",
                        help="""
                        Force the Exchange model to have the symmetry 
                        of the template.
                        """)

    return parser



r"""
Script core module contain the logic of the script functions and basically 
mirror the content of the scripts before the appearence of script-core, 
therefore the scripts from scripts folder are just the bindings to the 
corresponding files in script-core.
"""



r"""
Input-output from TB2J.
"""

import numpy as np

from rad_tools.exchange.model import ExchangeModel, Bond
from rad_tools.routines import absolute_to_relative


def read_exchange_model(filename, quiet=False) -> ExchangeModel:
    r"""
    Read exchange model from TB2J output file.

    Parameters
    ----------
    filename : str
        Path to the TB2J output file.
    quiet : bool, default True
        Whenever to supress output.

    Returns 
    -------
    model : :py:class:`.ExchangeModel`
        Exchange model buid from TB2J file.
    """

    major_sep = '=' * 90
    minor_sep = '-' * 88
    garbage = str.maketrans({'(': None,
                             ')': None,
                             '[': None,
                             ']': None,
                             ',': None,
                             '\'': None})
    cell_flag = 'Cell (Angstrom):'
    atoms_flag = 'Atoms:'
    atom_end_flag = 'Total'
    exchange_flag = 'Exchange:'
    iso_flag = 'J_iso:'
    aniso_flag = 'J_ani:'
    dmi_flag = 'DMI:'

    file = open(filename, "r")
    model = ExchangeModel()
    line = True
    atoms = {}

    # Read everything before exchange
    while line:
        line = file.readline()

        # Read cell
        if line and cell_flag in line:
            model.cell = np.array([
                list(map(float, file.readline().split())),
                list(map(float, file.readline().split())),
                list(map(float, file.readline().split()))
            ])

        # Read atoms
        if line and atoms_flag in line:
            line = file.readline()
            line = file.readline()
            line = file.readline().split()
            while line and atom_end_flag not in line:
                atoms[line[0]] = absolute_to_relative(
                    model.cell, *tuple(map(float, line[1:4])))
                line = file.readline().split()

        # Check if the exchange section is reached
        if line and exchange_flag in line:
            break

    # Read exchange
    while line:
        while line and minor_sep not in line:
            line = file.readline()
        line = file.readline().translate(garbage).split()
        atom1 = line[0]
        atom2 = line[1]
        R = tuple(map(int, line[2:5]))
        distance = float(line[-1])
        iso = None
        aniso = None
        dmi = None
        while line and minor_sep not in line:
            line = file.readline()

            # Read isotropic exchange
            if line and iso_flag in line:
                iso = float(line.split()[-1])

            # Read anisotropic exchange
            if line and aniso_flag in line:
                aniso = np.array([
                    list(map(float, file.readline().translate(garbage).split())),
                    list(map(float, file.readline().translate(garbage).split())),
                    list(map(float, file.readline().translate(garbage).split()))
                ])

            # Read DMI
            if line and dmi_flag in line:
                dmi = tuple(map(float, line.translate(garbage).split()[-3:]))

        # Adding info from the exchange block to the ExchangeModel structure
        if atom1 not in model.magnetic_atoms:
            model.add_atom(atom1, *atoms[atom1])
        if atom2 not in model.magnetic_atoms:
            model.add_atom(atom2, *atoms[atom2])
        bond = Bond(iso=iso, aniso=aniso, dmi=dmi)
        model.add_bond(bond, atom1, atom2, R)
        computed_distance = model.get_distance(atom1, atom2, R)
        if abs(computed_distance - distance) > 0.001 and not quiet:
            print(
                f"\nComputed distance is a different from the read one:\n  Computed: {computed_distance:.4f}\n  Read: {distance:.4f}\n")

    # Fill non-magnetic atoms
    for atom in atoms:
        if atom not in model.magnetic_atoms:
            model.nonmagnetic_atoms[atom] = np.array(atoms[atom])

    return model
